/*
 * Message parser for CMC RT-STAR GPS receiver
 * Author: Eric LaForest di458@freenet.carleton.ca
 * License: GPL
 * 
 * NOTE: endianness/IEEE754/signed-unsigned-char 
 * are the main factors that matter for porting.
 * This is very machine-specific code.
 */

#include "gpsparser.h"

struct rawmsgformat rxrawmsg, txrawmsg;		/* Raw CMC message from/to GPS receiver */
struct termios orig_values;	/* original port settings */
struct termios nmea_orig_values;/* original port settings for auxiliary */
int serialfd;			/* Serial port file descriptor */
int nmeafd;			/* Auxiliary output for NMEA-0183 format */

/* errno handling...these are generally fatal */
void crapout(char *ouch)
{
    perror(ouch);
    exit(1);
}

/* Only executed at program exit */
void resetserialport(void)
{
    if (tcsetattr(serialfd, TCSAFLUSH, &orig_values) == -1)
	crapout("resetserialport1");	/* Set port */

    if (tcsetattr(nmeafd, TCSAFLUSH, &nmea_orig_values) == -1)
	crapout("resetserialport1");	/* Set port */
}

int setserialport(char *port, char *speed)
{
    struct termios new_values;	/* desired serial port settings */
    struct termios test_values;	/* check for proper configuration */
    int ispeed;			/* desired port speed as int */
    speed_t bspeed;		/* desired port speed as B const */

    serialfd = open(port, O_RDWR);
    if (serialfd == -1)
	crapout(port);

    memset(&new_values, 0, sizeof(new_values));
    if (tcgetattr(serialfd, &orig_values) == -1)
	crapout(port);

    if (atexit(resetserialport) == -1) {
	fprintf(stderr, "Could not register atexit() function!!!\n");
	return -1;
    }
    /* Set a raw serial port */
    /* cfmakeraw(&new_values); */

    /*
     * These aim at having a completely non-canonical, dumb port.
     */
    new_values.c_cflag = CS8 | CLOCAL | CREAD;
    new_values.c_iflag = IGNPAR;
    new_values.c_oflag = 0;
    new_values.c_lflag = 0;
    new_values.c_cc[VTIME] = 0;
    new_values.c_cc[VMIN] = 1;

    ispeed = atoi(speed);
    switch (ispeed) {
    case 0:
	bspeed = B0;		/* Allowed, but is it useful? */
	break;
    case 50:
	bspeed = B50;
	break;
    case 75:
	bspeed = B75;
	break;
    case 110:
	bspeed = B110;
	break;
    case 134:
	bspeed = B134;
	break;
    case 150:
	bspeed = B150;
	break;
    case 200:
	bspeed = B200;
	break;
    case 300:
	bspeed = B300;
	break;
    case 600:
	bspeed = B600;
	break;
    case 1200:
	bspeed = B1200;
	break;
    case 1800:
	bspeed = B1800;
	break;
    case 2400:
	bspeed = B2400;
	break;
    case 4800:
	bspeed = B4800;
	break;
    case 9600:
	bspeed = B9600;
	break;
    case 19200:
	bspeed = B19200;
	break;
    case 38400:
	bspeed = B38400;
	break;
    case 57600:
	bspeed = B57600;
	break;
    case 115200:		/* 99.99% of cases can't go faster than this */
	bspeed = B115200;
	break;
    default:
	fprintf(stderr, "Invalid speed: %d\n", ispeed);
	return -1;
    }

    /* Set input/output speed */
    if (cfsetispeed(&new_values, bspeed) == -1)
	crapout(port);
    if (cfsetospeed(&new_values, bspeed) == -1)
	crapout(port);

    /* Set port and get back settings for checking */
    if (tcflush(serialfd, TCIFLUSH) == -1)
	crapout(port);
    if (tcsetattr(serialfd, TCSANOW, &new_values) == -1)
	crapout(port);
    if (tcgetattr(serialfd, &test_values) == -1)
	crapout(port);

    if (new_values.c_iflag != test_values.c_iflag) {
	fprintf(stderr, "c_iflag setting did not take\n");
	return -1;
    }
    if (new_values.c_oflag != test_values.c_oflag) {
	fprintf(stderr, "c_oflag setting did not take\n");
	return -1;
    }
    if (new_values.c_lflag != test_values.c_lflag) {
	fprintf(stderr, "c_lflag setting did not take\n");
	return -1;
    }
    if (new_values.c_cflag != test_values.c_cflag) {
	fprintf(stderr, "c_cflag setting did not take\n");
	return -1;
    }
    if (cfgetispeed(&test_values) != bspeed) {
	fprintf(stderr, "input speed did not take\n");
	return -1;
    }
    if (cfgetospeed(&test_values) != bspeed) {
	fprintf(stderr, "output speed did not take\n");
	return -1;
    }
    return 0;
}

int setserialport2(char *port, char *speed)
{
    struct termios new_values;	/* desired serial port settings */
    struct termios test_values;	/* check for proper configuration */
    int ispeed;			/* desired port speed as int */
    speed_t bspeed;		/* desired port speed as B const */

    nmeafd = open(port, O_RDWR);
    if (nmeafd == -1)
	crapout(port);

    memset(&new_values, 0, sizeof(new_values));
    if (tcgetattr(nmeafd, &orig_values) == -1)
	crapout(port);

    if (atexit(resetserialport) == -1) {
	fprintf(stderr, "Could not register atexit() function!!!\n");
	return -1;
    }
    /* Set a raw serial port */
    /* cfmakeraw(&new_values); */

    /*
     * These aim at having a completely non-canonical, dumb port.
     */
    new_values.c_cflag = CS8 | CLOCAL | CREAD;
    new_values.c_iflag = IGNPAR;
    new_values.c_oflag = 0;
    new_values.c_lflag = 0;
    new_values.c_cc[VTIME] = 0;
    new_values.c_cc[VMIN] = 1;

    ispeed = atoi(speed);
    switch (ispeed) {
    case 0:
	bspeed = B0;		/* Allowed, but is it useful? */
	break;
    case 50:
	bspeed = B50;
	break;
    case 75:
	bspeed = B75;
	break;
    case 110:
	bspeed = B110;
	break;
    case 134:
	bspeed = B134;
	break;
    case 150:
	bspeed = B150;
	break;
    case 200:
	bspeed = B200;
	break;
    case 300:
	bspeed = B300;
	break;
    case 600:
	bspeed = B600;
	break;
    case 1200:
	bspeed = B1200;
	break;
    case 1800:
	bspeed = B1800;
	break;
    case 2400:
	bspeed = B2400;
	break;
    case 4800:
	bspeed = B4800;
	break;
    case 9600:
	bspeed = B9600;
	break;
    case 19200:
	bspeed = B19200;
	break;
    case 38400:
	bspeed = B38400;
	break;
    case 57600:
	bspeed = B57600;
	break;
    case 115200:		/* 99.99% of cases can't go faster than this */
	bspeed = B115200;
	break;
    default:
	fprintf(stderr, "Invalid speed: %d\n", ispeed);
	return -1;
    }

    /* Set input/output speed */
    if (cfsetispeed(&new_values, bspeed) == -1)
	crapout(port);
    if (cfsetospeed(&new_values, bspeed) == -1)
	crapout(port);

    /* Set port and get back settings for checking */
    if (tcflush(nmeafd, TCIFLUSH) == -1)
	crapout(port);
    if (tcsetattr(nmeafd, TCSANOW, &new_values) == -1)
	crapout(port);
    if (tcgetattr(nmeafd, &test_values) == -1)
	crapout(port);

    if (new_values.c_iflag != test_values.c_iflag) {
	fprintf(stderr, "c_iflag setting did not take\n");
	return -1;
    }
    if (new_values.c_oflag != test_values.c_oflag) {
	fprintf(stderr, "c_oflag setting did not take\n");
	return -1;
    }
    if (new_values.c_lflag != test_values.c_lflag) {
	fprintf(stderr, "c_lflag setting did not take\n");
	return -1;
    }
    if (new_values.c_cflag != test_values.c_cflag) {
	fprintf(stderr, "c_cflag setting did not take\n");
	return -1;
    }
    if (cfgetispeed(&test_values) != bspeed) {
	fprintf(stderr, "input speed did not take\n");
	return -1;
    }
    if (cfgetospeed(&test_values) != bspeed) {
	fprintf(stderr, "output speed did not take\n");
	return -1;
    }
    return 0;
}

/* Returns when a message has been parsed succesfully or on error */
/* I'm too lazy to make a proper state machine... */
int find_message(void)
{
    int scratch;
    static int first = 1;	/* 1=first call 0=next calls */
    unsigned short checksum = 0;
    unsigned char buffer[3];	/* holds potential msg headers */
    unsigned char foo;

    /* Fill up the buffer first time around */
    if (first == 1) {
	/* fprintf(stderr, "Header buffer pre-fill...\n"); */
	if (read(serialfd, &buffer[0], 1) != 1) {
	    crapout("Error attempting to read msg header");
	}
	if (read(serialfd, &buffer[1], 1) != 1) {
	    crapout("Error attempting to read msg header");
	}
	if (read(serialfd, &buffer[2], 1) != 1) {
	    crapout("Error attempting to read msg header");
	}
	first = 0;		/* Never run this again */
    }
    while (1) {
	/* Find Start of Header char */
	/* fprintf(stderr, "Seeking for start-of-header..."); */
	buffer[0] = 0;		/* To clear out a past match */
	while (buffer[0] != 1) {
	    buffer[0] = buffer[1];
	    buffer[1] = buffer[2];
	    if (read(serialfd, &buffer[2], 1) != 1) {
		crapout("Error attempting to read msg header");
	    }
	    /* fprintf(stderr, "D=%02x", buffer[2]); */
	}

	/* Is this really a header? */
	foo = ~buffer[2];
	if (buffer[1] == foo) {
	    /* fprintf(stderr, "Yep, it's a header!\n"); */
	    rxrawmsg.start = buffer[0];
	    rxrawmsg.msgid = buffer[1];
	    rxrawmsg.msgidcmpl = buffer[2];
	    checksum += buffer[0] + buffer[1] + buffer[2];
	    /*
	     * fprintf(stderr, "Message Start: %02x\n", buffer[0]);
	     * fprintf(stderr, "Message ID: %02x\n", buffer[1]);
	     * fprintf(stderr, "Message ~ID: %02x\n", buffer[2]);
	     */
	    break;
	} else {
	    /* fprintf(stderr, "Nope, not a header...\007\n"); */
	    /*
	     * fprintf(stderr, "%02x %02x %02x %02x %02x %d\n", buffer[0], buffer[1], buffer[2], ~buffer[1], ~buffer[2], (buffer[1] == ~(buffer[2])) );
	     */
	    continue;
	}
    }

    /* This far means a correct message ID# header */
    /* fprintf(stderr, "Reading message length..."); */
    if (read(serialfd, &rxrawmsg.msglen, 1) != 1) {
	crapout("Error reading message length");
    }
    /* fprintf(stderr, "%02x\n", rxrawmsg.msglen); */
    checksum += rxrawmsg.msglen;

    /* Read in the actual message */
    if (rxrawmsg.msglen != 0) {
	/* fprintf(stderr, "Reading message...\n"); */
	for (scratch = 0; scratch < rxrawmsg.msglen; scratch++) {
	    if (read(serialfd, &rxrawmsg.message[scratch], 1) != 1) {
		crapout("Error reading message data");
	    }
	    checksum += rxrawmsg.message[scratch];
	}
    }
    /* If the checksum fails, drop it all */
    /* fprintf(stderr, "Getting checksum..."); */
    if (read(serialfd, &rxrawmsg.checksumlsb, 1) != 1) {
	crapout("Error reading message checksum LSB");
    }
    if (read(serialfd, &rxrawmsg.checksummsb, 1) != 1) {
	crapout("Error reading message checksum MSB");
    }
    /* fprintf(stderr, "%02x%02x\n", rxrawmsg.checksummsb, rxrawmsg.checksumlsb); */
    if (checksum != (unsigned short) ((rxrawmsg.checksummsb << 8) | rxrawmsg.checksumlsb)) {
	/* fprintf(stderr, "Bad message checksum...C:%04x R:%04x\n", checksum, (unsigned short)((rxrawmsg.checksummsb<<8) | rxrawmsg.checksumlsb)); */
	return -1;
    }
    /* fprintf(stderr, "Checksum OK...this is a live one\n"); */
    /* Got this far: message should be correct and ready to parse */

    return 0;
}

/*
 * Parse incoming messages
 */
void parse_rx_msg(void)
{
    if (find_message() == 0) {	/* Find and store message */
	switch (rxrawmsg.msgid) {	/* Call the apropriate parser */

	case 20:	/* Navigation result */
	    rxparser20(rxrawmsg);
	    break;

	case 6:
	    rxparser6(rxrawmsg);
	    break;
	
	case 7:
	    rxparser7(rxrawmsg);
	    break;

	case 63:
	    break;	/* Startup message */

	case 126:	/* Acknowledge message */
	    break;

	default:
//	    status ("Received message %d!", rxrawmsg.msgid);
//	    refresh();
	    break;
	}
    }
}

/*
 * Converts a hex digit into its decimal value
 * Quick and dirty --- No error checking!!!
 * Symbols in the gaps between the acceptable ranges
 * will yield bogus results
 */
unsigned char ctoh(unsigned char hexdig)
{
    if (hexdig <= 57) {		/* [0-9] */
	return (hexdig - 48);
    }
    if (hexdig <= 70) {		/* [A-F] */
	return (hexdig - 55);
    }
    if (hexdig <= 102) {	/* [a-f] */
	return (hexdig - 87);
    }
    return 0;			/* any ASCII symbol above 102 */
}
